/*
 * QIC cartridge tape drive, SCSI controller. For use with Caliper, Sankyo,
 * Wangtek.
 * 
 * ENOSPC is returned at end-of-tape.  EIO is returned by Media Error, but read
 * resync is possible.  New Sankyo CP-150SE supported.
 *
 * Changes for the Archive Viper Copyright (c) 1991 by Frank J. Edwards.
 * All other code property of Commodore-Amiga, Inc.
 */

#include    "sys/types.h"
#include    "sys/param.h"
#include    "sys/buf.h"
#include    "sys/inline.h"
#include    "sys/errno.h"
#include    "sys/file.h"
#include    "sys/uio.h"
#include    "rico.h"
#include    "sd.h"

#include    "viper.h"           /* should be "sys/tapeio.h" */

/*
 * minor device options MUST be coordinated with `sdunit' and `sdcard'
 * macros.
 */
#define ctrewind(d)     (not ((d) & 1<<7))
#define ctappend(d)     ((d) & 1<<6)
#define ctden150(d)     ((d) & 1<<5)

struct ct {
    uint            state;
    bool            open;
    struct buf     *bhead, *btail;
    struct sdcom    com;
    uchar           sense[14];
    ulong           debug;      /* New field as compared to ct.c */
};

/*
 * ct.state
 */
#define NORMAL_IDLE             0
#define NORMAL_READ             1
#define NORMAL_SPECIAL          2
#define NORMAL_SEOF             3
#define NORMAL_READ_GETSENSE    4
#define NORMAL_SEOF_GETSENSE    5
#define EOD_IDLE                6
#define EOD_WRITE               7
#define EOD_WRITE_GETSENSE      8
#define EOD_SPECIAL             9
#define EOM_IDLE                10
#define EOM_SPECIAL             11
#define EOF_IDLE                12
#define EOF_SPECIAL             13
#define DOOM_IDLE               14
#define DOOM_SPECIAL            15
#define DOOM_MISC_GETSENSE      16
#define BAD_READ_1              17
#define BAD_READ_2              18
#define BAD_READ_NULS           19

#define STATUS_INQUIRE          20  /* Really not a _state_ but a cmd */
#define STATUS_MODE_SENSE       21  /* ditto */
#define STATUS_MODE_SELECT      22
#define STATUS_REQ_SENSE        23

struct special {
    bool            active;
    uchar           com, sense[sizeof(((struct ct *) 0)->sense)];
    struct buf      buf;
};

/*
 * special.com
 */
#define INIT120     0
#define INIT150     1
#define REWIND      2
#define GETSENSE    3
#define APPEND      4
#define WEOF        5
#define FLUSH       6
#define SEOF        7
#define IO          8
#define CLOSE       9

#define VIPER       2150        /* Model number supported */
#define BSIZE       512
#define AT_BOT(x)   /* nothing -- for now */

static struct ct cttab[SDCARDS][SDUNITS];
static struct special special;

int             ctopen(dev_t * devp, int flags);
int             ctclose(dev_t dev, int flags);
int             ctread(dev_t dev, struct uio * uiop);
int             ctwrite(dev_t dev, struct uio * uiop);

static void     dump(uchar * buf, long len);
static void     dump_sdcom(char *msg, struct sdcom * cp);
static char    *sensekey(int sense);
static void     dump_sense(char *file, int line, uchar * sense);
static void     breakup(struct buf * bp);
static void     strategy(struct buf * bp);
static int      init(dev_t dev);
static int      command(struct ct * dp, int com);
static int      queue(struct ct * dp, struct buf * bp);
static void     ihandle(struct sdcom * cp);
static int      machine(struct ct * dp);
static int      dispose(struct ct * dp, int st);
static int      start(struct ct * dp, int com, int st);
static int      setresidue(uchar sense[], struct buf * bp);
static int      report(struct ct * dp);
static int      printbits(int bits, char *names);

#ifndef DEBUG
# define DEBUG_ON(x)    if (0) {
# define DEBUG_OFF(x)   }
#else
# define DEBUG_ON(x)    if (dp->debug & (x)) {
# define DEBUG_OFF(x)   }

# define STATE_DATA     /* nothing */
# define FILEMARK       0x80
# define EOM            0x40
# define SENSEKEY       0x0F

/* Called to display sense data */
static void     dump(uchar * buf, long len)
{
    if (len > 0) {
        register int    pos = 0;
        while (len-- > 0) {
            printf("%x ", *buf++);
            pos += 3;
            if (pos > 76) {
                putchar('\n');
                pos = 0;
            }
        }
        if (pos)
            putchar('\n');
    }
}

static void     dump_sdcom(char *msg, struct sdcom * cp)
{
    printf("Contents of (struct sdcom)%s:\n", msg);
    printf(" sdcom.okay:  %x    sdcom.status:  %x\n", cp->okay, cp->status);
    printf(" sdcom.nbyte: %x    sdcom.cdb:     ", cp->nbyte);
    dump((uchar *) cp->cdb, 6);
}

static void     dump_sense(char *file, int line, uchar * sense)
{
    printf("Sense data from GETSENSE @ %s:%d\n ", file, line);
    dump(sense, 14);
    printf(" %s%s%s\n", sense[2] & FILEMARK ? "Filemark, " : "",
        sense[2] & EOM ? "EOM, " : "",
        sensekey(sense[2] & SENSEKEY));
}
#endif              /* DEBUG */

static char    *sensekey(int sense)
{
    static char    *SenseKey[] = {
        "FM/EOM or no status available",
        "Recovered from error; command successful",
        "Not Ready",
        "Medium Error",
        "Hardware Error",
        "Illegal Request (illegal parameter in CDB)",
        "Unit Attention (tape changed or Viper reset)",
        "Data Protect (cartridge is write-portected)",
        "Blank Check (no-data condition found on tape)",
        "Vendor Unique Code -- not used on Viper",
        "-- undocumented error --",
        "Aborted Command (may retry)",
        "-- undocumented error --",
        "Volume Overflow (drive buffer may contain data)",
    };
#define KEY_SIZE    sizeof(SenseKey)
    return (SenseKey[sense]);
}

#ifdef STATE_DATA

#define STRING(x)   "" # x

static char *States[] = {
    STRING( NORMAL_IDLE ),
    STRING( NORMAL_READ ),
    STRING( NORMAL_SPECIAL ),
    STRING( NORMAL_SEOF ),
    STRING( NORMAL_READ_GETSENSE ),
    STRING( NORMAL_SEOF_GETSENSE ),
    STRING( EOD_IDLE ),
    STRING( EOD_WRITE ),
    STRING( EOD_WRITE_GETSENSE ),
    STRING( EOD_SPECIAL ),
    STRING( EOM_IDLE ),
    STRING( EOM_SPECIAL ),
    STRING( EOF_IDLE ),
    STRING( EOF_SPECIAL ),
    STRING( DOOM_IDLE ),
    STRING( DOOM_SPECIAL ),
    STRING( DOOM_MISC_GETSENSE ),
    STRING( BAD_READ_1 ),
    STRING( BAD_READ_2 ),
    STRING( BAD_READ_NULS ),
    STRING( STATUS_INQUIRE ),
    STRING( STATUS_MODE_SENSE ),
    STRING( STATUS_MODE_SELECT ),
    STRING( STATUS_REQ_SENSE ),
};
static char *Commands[] = {
    STRING( INIT120 ),
    STRING( INIT150 ),
    STRING( REWIND ),
    STRING( GETSENSE ),
    STRING( APPEND ),
    STRING( WEOF ),
    STRING( FLUSH ),
    STRING( SEOF ),
    STRING( IO ),
    STRING( CLOSE ),
};

#undef STRING

#endif /* STATE_DATA */

int ctioctl(dev_t *devp, int cmd, caddr_t arg)
{
    struct ct *dp = &cttab [sdcard(*devp)] [sdunit(*devp)] ;
    struct vio_arg *varg;
    int state, err, size;

    switch (cmd) {
    case VIODBGON :
        err = dp->debug;
        if ((long)arg & ~VIO_ALL)       /* Unused bits were set! */
            return( EINVAL );
        dp->debug |= (ulong) arg;
        break;
    case VIODBGOFF :
        err = dp->debug;
        if ((long)arg & ~VIO_ALL)       /* Unused bits were set! */
            return( EINVAL );
        dp->debug &= ~ (ulong) arg;
        break;

    default :
        err = 0;
        size = 0;
        varg = (struct vio_arg *)arg;
        if (arg) {
            state = dp->state;
            switch (cmd) {
            case VIOINQ :   cmd = STATUS_INQUIRE;       size = 36;  break;
            case VIOMSENS : cmd = STATUS_MODE_SENSE;    size = 12;  break;
            case VIORSENS : cmd = STATUS_REQ_SENSE;     size = 14;  break;
            case VIOMSEL :  cmd = STATUS_MODE_SELECT;   size = 12;
                            copyin(varg->vio_addr, dp->sense, 12);  break;
            }
            err = command(dp, cmd);
            dp->state = state;
            if (!err && copyout(dp->sense, varg->vio_addr, size))
                err = EFAULT, size = 0;
        }
        varg->vio_len = size;
    }
    return( err );
}

int ctopen(dev_t * devp, int flags)
{
    struct ct      *dp;
    uchar           sense2, sense8, sense9;
    int             e;

    /* Error if the requested device is already open */
    dp = &cttab[sdcard(*devp)][sdunit(*devp)];
    if (dp->open)
        return (EBUSY);

    /* Determine accessibility of SCSI ID */
    if ((e = sdopen(sdcard(*devp))) or (e = command(dp, GETSENSE)))
        return (e);

    sense2 = dp->sense[2];
    sense8 = dp->sense[8];
    sense9 = dp->sense[9];
DEBUG_ON( VIO_GENERAL )
        dump_sdcom(" after sdopen in ctopen()", &dp->com);
        dump_sense(__FILE__, __LINE__, dp->sense);
DEBUG_OFF( VIO_GENERAL )

    sense2 = dp->state;         /* Have to save the current state */
    e = command(dp, STATUS_MODE_SENSE);
    dp->state = sense2;         /*  ..and restore it later */
    if (e)
        return (e);

    if (dp->sense[4]) {
        /* The tape already been set to a particular density... */
        if (dp->sense[4] != (ctden150(*devp) ? 0x10 : 0x0F)) {
            printf("%s: tape ct%d: ", sdhardwarename(dp->com.card),
               dp->com.card * SDUNITS + dp->com.unit);
            printf("Cannot change tape density on-the-fly\n");
            return( EIO );
        }
    } else /* if (((FREAD | FWRITE) & flags) == FWRITE) */ {
        /* The tape either has not been used or has been reset... */
        dp->state = NORMAL_IDLE;
        if (e = init(*devp))
            return (e);
    }
    switch (flags & (FREAD | FWRITE)) {
    case FREAD:
        if (ctappend(*devp))
            return (EINVAL);
        break;
    case FWRITE:
        if ((flags & FAPPEND) or (ctappend(*devp)))
            if (e = command(dp, APPEND))
                return (e);
        break;
    default:
DEBUG_ON( VIO_DEFAULT )
        printf("strange mode in ctopen(): %x\n", flags);
DEBUG_OFF( VIO_DEFAULT )
        return (EINVAL);
    }
    dp->open = TRUE;
    return (0);
}

int ctclose(dev_t dev, int flags)
{
    struct ct      *dp;
    dp = &cttab[sdcard(dev)][sdunit(dev)];
    if (ctrewind(dev)) {
        if (flags & FWRITE)
            command(dp, WEOF);
        command(dp, REWIND);
    } else {
        if (flags & FWRITE)
            command(dp, WEOF);
        else
            command(dp, SEOF);
    }
    command(dp, CLOSE);
    dp->open = FALSE;
    return (0);
}

int ctread(dev_t dev, struct uio * uiop)
{
    return (uiophysio(breakup, (struct buf *) 0, dev, B_READ, uiop));
}

int ctwrite(dev_t dev, struct uio * uiop)
{
    return (uiophysio(breakup, (struct buf *) 0, dev, B_WRITE, uiop));
}

static void     breakup(struct buf * bp)
{
    amiga_dma_pageio(strategy, bp);
}

static void     strategy(struct buf * bp)
{
    int             x;
    bp->b_resid = 0;
    x = sdspl();
    queue(&cttab[sdcard(bp->b_edev)][sdunit(bp->b_edev)], bp);
    splx(x);
}

static int      init(dev_t dev)
{
    int             com = ctden150(dev) ? INIT150 : INIT120;
    return (command(&cttab[sdcard(dev)][sdunit(dev)], com) ? ENXIO : 0);
}

static int      command(struct ct * dp, int com)
{
    static bool     locked;
    int             x;

    x = sdspl();
DEBUG_ON( VIO_GENERAL )
    dump_sdcom(" at beginning of command()", &dp->com);
DEBUG_OFF( VIO_GENERAL )
    while (locked)
        sleep(&locked, PZERO);
    locked = TRUE;
    special.com = com;
    special.buf.b_flags = 0;
    special.active = TRUE;
    queue(dp, &special.buf);
    while (special.active)
        sleep(&special.active, PZERO);
    splx(x);
    locked = FALSE;
    wakeup(&locked);

    /* Shouldn't the value of special be checked *before* the wakeup()? */
    if (special.buf.b_flags & B_ERROR) {
DEBUG_ON( VIO_ERROR )
DEBUG_ON( VIO_GENERAL )
        dump_sense(__FILE__, __LINE__, dp->sense);
DEBUG_OFF( VIO_GENERAL )
        printf("Error %d (flags 0x%x) in CT:command(..., %d)\n",
               special.buf.b_error, special.buf.b_flags, com);
DEBUG_OFF( VIO_ERROR )
        return (special.buf.b_error ? special.buf.b_error : EIO);
    }
    return (0);
}

static int      queue(struct ct * dp, struct buf * bp)
{
    bp->av_forw = 0;
    if (dp->bhead) {
        dp->btail->av_forw = bp;
        dp->btail = bp;
    } else {
        dp->bhead = bp;
        dp->btail = bp;
        machine(dp);
    }
}

static void     ihandle(struct sdcom * cp)
{
    machine(&cttab[cp->card][cp->unit]);
}

static int      machine(struct ct * dp)
{
    struct buf     *bp;
    while (bp = dp->bhead) {

DEBUG_ON( VIO_STATEDATA )
        if (bp == &special.buf) {
            printf("Entering state %d (%s) with command %d (%s)\n",
                dp->state, States[dp->state],
                special.com, Commands[special.com]);
        } else {
            printf("Entering state %d (%s) for function %s\n",
                dp->state, States[dp->state],
                bp->b_flags & B_READ ? "B_READ" : "B_WRITE");
        }
DEBUG_OFF( VIO_STATEDATA )

        switch (dp->state) {
        case STATUS_INQUIRE:
        case STATUS_MODE_SENSE:
        case STATUS_MODE_SELECT:
            start(dp, dp->state, dp->state);    /* Infinite loop? ;-) */
            return;
        case STATUS_REQ_SENSE:
            start(dp, GETSENSE, dp->state);
            return;

        case NORMAL_IDLE:
            unless(bp == &special.buf) {
                if (bp->b_flags & B_READ)
                    start(dp, IO, NORMAL_READ);
                else
                    start(dp, IO, EOD_WRITE);
                return;
            }
            switch (special.com) {
            case INIT120:
            case INIT150:
            case WEOF:
                start(dp, special.com, NORMAL_SPECIAL);
                return;
            case GETSENSE:
                start(dp, special.com, NORMAL_SPECIAL);
DEBUG_ON( VIO_GETSENSE )
                dump_sense(__FILE__, __LINE__, dp->sense);
DEBUG_OFF( VIO_GETSENSE )
                return;
            case REWIND:
            /*  start(dp, special.com, DOOM_SPECIAL);   /* */
                start(dp, special.com, NORMAL_SPECIAL); /* [FJE] */
                return;
            case APPEND:
                start(dp, special.com, EOD_SPECIAL);
                return;
            case SEOF:
                start(dp, special.com, NORMAL_SEOF);
                return;
            default:
DEBUG_ON( VIO_DEFAULT )
                printf(" ignored special.com:  %x\n", special.com);
DEBUG_OFF( VIO_DEFAULT )
                dispose(dp, NORMAL_IDLE);
            }
            break;
        case NORMAL_READ:
            if (!dp->com.okay)
                dp->state = DOOM_MISC_GETSENSE;
            else if (dp->com.status == 0)
                dispose(dp, NORMAL_IDLE);
            else if (dp->com.status == 2) {
                start(dp, GETSENSE, NORMAL_READ_GETSENSE);
                return;
            } else
                dp->state = DOOM_MISC_GETSENSE;
            break;

        case NORMAL_READ_GETSENSE:
            if ((not dp->com.okay) or(dp->com.status != 0)) {
                dp->state = DOOM_MISC_GETSENSE;
                break;
            }
DEBUG_ON( VIO_GENERAL )
            dump_sense(__FILE__, __LINE__, dp->sense);
DEBUG_OFF( VIO_GENERAL )
            switch (dp->sense[2] & 0x0F) {
            case 0:         /* No sense data available */
                if (setresidue(dp->sense, bp) < bp->b_bcount)
                    dispose(dp, EOF_IDLE);
                dp->state = EOF_IDLE;
                break;
            case 8:         /* Blank Check */
                report(dp);
                if (setresidue(dp->sense, bp) < bp->b_bcount)
                    dispose(dp, EOD_IDLE);
                dp->state = EOD_IDLE;
                break;
            case 3:         /* Media Error */
                report(dp);
                if (setresidue(dp->sense, bp) < bp->b_bcount)
                    dispose(dp, BAD_READ_1);
                dp->state = BAD_READ_1;
                break;
            default:
                report(dp);
DEBUG_ON( VIO_DEFAULT )
                printf(" ignored sense[2]:  %x\n", dp->sense[2] & 0x0F);
                dump_sense(__FILE__, __LINE__, dp->sense);
DEBUG_OFF( VIO_DEFAULT )
                dp->state = DOOM_IDLE;
            }
            break;
        case BAD_READ_1:
            if (bp == &special.buf)
                dp->state = NORMAL_IDLE;
            else {
DEBUG_ON( VIO_GENERAL )
                printf("Setting B_ERROR @ %s:%d\n", __FILE__, __LINE__);
DEBUG_OFF( VIO_GENERAL )
                bp->b_flags |= B_ERROR;
                dispose(dp, BAD_READ_2);
            }
            break;

        case BAD_READ_2:
            if (bp == &special.buf)
                dp->state = NORMAL_IDLE;
            else {
                 /* [FJE] */ bp->b_error = EIO;
                bp->b_flags |= B_ERROR;
DEBUG_ON( VIO_GENERAL )
                printf("Setting B_ERROR @ %s:%d\n", __FILE__, __LINE__);
DEBUG_OFF( VIO_GENERAL )
                dispose(dp, BAD_READ_NULS);
            }
            break;
        case BAD_READ_NULS:
            unless(bp == &special.buf) {
                bzero(vtop(paddr(bp), bp->b_proc), BSIZE);
                bp->b_resid = bp->b_bcount - BSIZE;
                dispose(dp, NORMAL_IDLE);
            }
            dp->state = NORMAL_IDLE;
            break;
        case NORMAL_SPECIAL:
            if (!dp->com.okay)
                dp->state = DOOM_MISC_GETSENSE;
            else if (dp->com.status == 0)
                dispose(dp, NORMAL_IDLE);
            else if (dp->com.status == 2) {
                start(dp, GETSENSE, DOOM_MISC_GETSENSE);
                return;
            } else
                dp->state = DOOM_MISC_GETSENSE;
            break;
        case NORMAL_SEOF:
            if (!dp->com.okay)
                dp->state = DOOM_MISC_GETSENSE;
            else if (dp->com.status == 0)
                dispose(dp, EOF_IDLE);
            else if (dp->com.status == 2) {
                start(dp, GETSENSE, NORMAL_SEOF_GETSENSE);
                return;
            } else
                dp->state = DOOM_MISC_GETSENSE;
            break;
        case NORMAL_SEOF_GETSENSE:
            if ((dp->com.okay) and(dp->com.status == 0)
                and((dp->sense[2] & 0x0F) == 8)) {
                report(dp);
                dispose(dp, EOD_IDLE);
            } else
                dp->state = DOOM_MISC_GETSENSE;
            break;

        case EOD_IDLE:
            unless(bp == &special.buf) {
                if (bp->b_flags & B_READ) {
                    bp->b_error = ENOSPC;
                    bp->b_flags |= B_ERROR;
                    dispose(dp, EOD_IDLE);
                    break;
                }
                start(dp, IO, EOD_WRITE);
                return;
            }
            switch (special.com) {
            case REWIND:
            /*  start(dp, special.com, DOOM_SPECIAL);   /* */
                start(dp, special.com, NORMAL_SPECIAL); /* [FJE] */
                return;
            case GETSENSE:
                start(dp, special.com, EOD_SPECIAL);
DEBUG_ON( VIO_GETSENSE )
                dump_sense(__FILE__, __LINE__, dp->sense);
DEBUG_OFF( VIO_GETSENSE )
                return;
            case WEOF:
                start(dp, special.com, EOD_SPECIAL);
                return;
            default:
DEBUG_ON( VIO_DEFAULT )
                printf(" ignored special.com:  %x\n", special.com);
DEBUG_OFF( VIO_DEFAULT )
                dispose(dp, EOD_IDLE);
            }
            break;
        case EOD_WRITE:
            if (!dp->com.okay)
                dp->state = DOOM_MISC_GETSENSE;
            else if (dp->com.status == 0)
                dispose(dp, EOD_IDLE);
            else if (dp->com.status == 2) {
                start(dp, GETSENSE, EOD_WRITE_GETSENSE);
                return;
            } else
                dp->state = DOOM_MISC_GETSENSE;
            break;

        case EOD_WRITE_GETSENSE:
            if ((not dp->com.okay) or(dp->com.status != 0)) {
                dp->state = DOOM_MISC_GETSENSE;
                break;
            }
DEBUG_ON( VIO_GENERAL )
            dump_sense(__FILE__, __LINE__, dp->sense);
DEBUG_OFF( VIO_GENERAL )
            switch (dp->sense[2] & 0xF) {
            case 0x0:   /* No sense data available */
            case 0xD:   /* Volume overflow */
                report(dp);
                if (setresidue(dp->sense, bp) < bp->b_bcount)
                    dispose(dp, EOM_IDLE);
                dp->state = EOM_IDLE;
                break;
            default:
                report(dp);
DEBUG_ON( VIO_DEFAULT )
                printf(" ignored sense[2]:  %x\n", dp->sense[2] & 0x0F);
DEBUG_OFF( VIO_DEFAULT )
                dp->state = DOOM_IDLE;
            }
            break;
        case EOD_SPECIAL:
            if (!dp->com.okay)
                dp->state = DOOM_MISC_GETSENSE;
            else if (dp->com.status == 0)
                dispose(dp, EOD_IDLE);
            else if (dp->com.status == 2) {
                start(dp, GETSENSE, DOOM_MISC_GETSENSE);
                return;
            } else
                dp->state = DOOM_MISC_GETSENSE;
            break;
        case EOM_IDLE:
            unless(bp == &special.buf) {
                bp->b_error = ENOSPC;
                bp->b_flags |= B_ERROR;
                dispose(dp, EOM_IDLE);
                break;
            }

            switch (special.com) {
            case REWIND:
            /*  start(dp, special.com, DOOM_SPECIAL);   /* */
                start(dp, special.com, NORMAL_SPECIAL); /* [FJE] */
                return;
            case GETSENSE:
                start(dp, special.com, EOM_SPECIAL);
DEBUG_ON( VIO_GETSENSE )
                dump_sense(__FILE__, __LINE__, dp->sense);
DEBUG_OFF( VIO_GETSENSE )
                return;
            default:
DEBUG_ON( VIO_DEFAULT )
                printf(" ignored special.com:  %x\n", special.com);
DEBUG_OFF( VIO_DEFAULT )
                dispose(dp, EOM_IDLE);
            }
            break;
        case EOM_SPECIAL:
            if (!dp->com.okay)
                dp->state = DOOM_MISC_GETSENSE;
            else if (dp->com.status == 0)
                dispose(dp, EOM_IDLE);
            else if (dp->com.status == 2) {
                start(dp, GETSENSE, DOOM_MISC_GETSENSE);
                return;
            } else
                dp->state = DOOM_MISC_GETSENSE;
            break;
        case EOF_IDLE:
            unless(bp == &special.buf) {
                bp->b_resid = bp->b_bcount;
                dispose(dp, EOF_IDLE);
                break;
            }
            switch (special.com) {
            case REWIND:
            /*  start(dp, special.com, DOOM_SPECIAL);   /* */
                start(dp, special.com, NORMAL_SPECIAL); /* [FJE] */
                return;
            case GETSENSE:
                start(dp, special.com, EOF_SPECIAL);
DEBUG_ON( VIO_GETSENSE )
                dump_sense(__FILE__, __LINE__, dp->sense);
DEBUG_OFF( VIO_GETSENSE )
                return;
            case CLOSE:
                dispose(dp, NORMAL_IDLE);
                break;
            default:
DEBUG_ON( VIO_DEFAULT )
                printf(" ignored special.com:  %x\n", special.com);
DEBUG_OFF( VIO_DEFAULT )
                dispose(dp, EOF_IDLE);
            }
            break;

        case EOF_SPECIAL:
            if (!dp->com.okay)
                dp->state = DOOM_MISC_GETSENSE;
            else if (dp->com.status == 0)
                dispose(dp, EOF_IDLE);
            else if (dp->com.status == 2) {
                start(dp, GETSENSE, DOOM_MISC_GETSENSE);
                return;
            } else
                dp->state = DOOM_MISC_GETSENSE;
            break;
        case DOOM_IDLE:
            unless(bp == &special.buf) {
DEBUG_ON( VIO_GENERAL )
                printf("Setting B_ERROR @ %s:%d\n", __FILE__, __LINE__);
DEBUG_OFF( VIO_GENERAL )
                bp->b_flags |= B_ERROR; /* */
                dispose(dp, DOOM_IDLE);
                break;
            }
            switch (special.com) {
            case REWIND:
            /*  start(dp, special.com, DOOM_SPECIAL);   /* */
                start(dp, special.com, NORMAL_SPECIAL); /* [FJE] */
                return;
            case GETSENSE:
            /*  start(dp, special.com, DOOM_SPECIAL);   /* */
                start(dp, special.com, NORMAL_SPECIAL); /* [FJE] */
DEBUG_ON( VIO_GETSENSE )
                dump_sense(__FILE__, __LINE__, dp->sense);
DEBUG_OFF( VIO_GETSENSE )
                return;
            default:
DEBUG_ON( VIO_DEFAULT )
                printf("Setting B_ERROR @ %s:%d\n", __FILE__, __LINE__);
                printf(" error from special.com:  %x\n", special.com);
                dump_sdcom(" in DOOM_IDLE of start", &dp->com);
DEBUG_OFF( VIO_DEFAULT )
                bp->b_flags |= B_ERROR; /* */
                dispose(dp, DOOM_IDLE);
            }
            break;
        case DOOM_SPECIAL:
            if (!dp->com.okay)
                dp->state = DOOM_MISC_GETSENSE;
            else if (dp->com.status == 0)
                dispose(dp, DOOM_IDLE);
            else if (dp->com.status == 2) {
                start(dp, GETSENSE, DOOM_MISC_GETSENSE);
                return;
            } else
                dp->state = DOOM_MISC_GETSENSE;
            break;

        case DOOM_MISC_GETSENSE:
            report(dp);
DEBUG_ON( VIO_GENERAL )
            printf("Setting B_ERROR @ %s:%d (DOOM_MISC_GETSENSE)\n",
                __FILE__, __LINE__);
            dump_sense(__FILE__, __LINE__, dp->sense);
DEBUG_OFF( VIO_GENERAL )
            bp->b_flags |= B_ERROR; /* */
            dispose(dp, DOOM_IDLE);
        }
    }
}

static int      dispose(struct ct * dp, int st)
{
    struct buf     *bp;
    bp = dp->bhead;
    dp->bhead = bp->av_forw;
    if (bp == &special.buf) {
        bcopy(dp->sense, special.sense, sizeof special.sense);
        special.active = FALSE;
        wakeup(&special.active);
    } else
        iodone(bp);
    dp->state = st;
}

static int      start(struct ct * dp, int com, int st)
{
    static uchar    init120[] = {
        0x00, 0x00, 0x10, 0x08,
        0x0F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x00
    /*  ^^^^  note the value "15" */
    };
    static uchar    init150[] = {
        0x00, 0x00, 0x10, 0x08,
        0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x00
    /*  ^^^^  note the value "16" */
    };
    struct buf     *bp;

    dp->com.nbyte = 0;
    switch (com) {
    case APPEND:
        AT_BOT(FALSE);
        dp->com.cdb[0] = 0x11;  /* (O) Space */
        dp->com.cdb[1] = 0x03;  /* Move to EOD */
        dp->com.cdb[2] = 0;
        dp->com.cdb[3] = 0;
        dp->com.cdb[4] = 0;
        dp->com.cdb[5] = 0;
        dp->com.reading = FALSE;
        break;
/*  case STATUS_REQ_SENSE:  /* Uses code for GETSENSE */
    case STATUS_INQUIRE:
    case STATUS_MODE_SENSE:
    case STATUS_MODE_SELECT:
        switch (com) {
        dp->com.reading = TRUE;
        case STATUS_INQUIRE:
            dp->com.cdb[0] = 0x12;  /* (O) Mode Sense */
            dp->com.nbyte = 36;     /* Viper returns 12 bytes of mode-sense */
            break;
        case STATUS_MODE_SENSE:
            dp->com.cdb[0] = 0x1a;  /* (O) Mode Sense */
            dp->com.nbyte = 12;     /* Viper returns 12 bytes of mode-sense */
            break;
        case STATUS_MODE_SELECT:
            dp->com.cdb[0] = 0x15;  /* (O) Mode Sense */
            dp->com.nbyte = 12;     /* Viper returns 12 bytes of mode-sense */
            dp->com.reading = FALSE;
            break;
        }
        dp->com.addr = (caddr_t) dp->sense;
        dp->com.cdb[1] = 0;
        dp->com.cdb[2] = 0;
        dp->com.cdb[3] = 0;
        dp->com.cdb[4] = dp->com.nbyte;
        dp->com.cdb[5] = 0;
        break;
    case GETSENSE:
        dp->com.addr = (caddr_t) dp->sense;
        dp->com.nbyte = sizeof dp->sense;
        dp->com.cdb[0] = 0x03;  /* (M) Request Sense */
        dp->com.cdb[1] = 0;
        dp->com.cdb[2] = 0;
        dp->com.cdb[3] = 0;
        dp->com.cdb[4] = sizeof dp->sense;
        dp->com.cdb[5] = 0;
        dp->com.reading = TRUE;
        break;
    case INIT120:
        dp->com.addr = (caddr_t) init120;
        dp->com.nbyte = sizeof init120;
        dp->com.cdb[0] = 0x15;  /* (O) Mode Select */
        dp->com.cdb[1] = 0;
        dp->com.cdb[2] = 0;
        dp->com.cdb[3] = 0;
        dp->com.cdb[4] = 0x0C;
        dp->com.cdb[5] = 0;
        dp->com.reading = FALSE;
        break;
    case INIT150:
        dp->com.addr = (caddr_t) init150;
        dp->com.nbyte = sizeof init150;
        dp->com.cdb[0] = 0x15;  /* (O) Mode Select */
        dp->com.cdb[1] = 0;
        dp->com.cdb[2] = 0;
        dp->com.cdb[3] = 0;
        dp->com.cdb[4] = 0x0C;
        dp->com.cdb[5] = 0;
        dp->com.reading = FALSE;
        break;
    case IO:
        AT_BOT(FALSE);
        bp = dp->bhead;
        dp->com.addr = (caddr_t)
            vtop(paddr(bp), bp->b_flags & B_KERNBUF ? 0 : bp->b_proc);
        dp->com.nbyte = bp->b_bcount;
        if (bp->b_flags & B_READ) {
            dp->com.reading = TRUE;
            dp->com.cdb[0] = 0x08;  /* (M) Read */
        } else {
            dp->com.reading = FALSE;
            dp->com.cdb[0] = 0x0A;  /* (M) Write */
        }
        dp->com.cdb[1] = 0x01;
        dp->com.cdb[2] = byte2(bp->b_bcount / BSIZE);
        dp->com.cdb[3] = byte1(bp->b_bcount / BSIZE);
        dp->com.cdb[4] = byte0(bp->b_bcount / BSIZE);
        dp->com.cdb[5] = 0;
        break;
    case REWIND:
        AT_BOT(TRUE);
        dp->com.cdb[0] = 0x01;  /* (M) Rewind */
        dp->com.cdb[1] = 0;
        dp->com.cdb[2] = 0;
        dp->com.cdb[3] = 0;
        dp->com.cdb[4] = 0;
        dp->com.cdb[5] = 0;
        dp->com.reading = FALSE;
        break;
    case SEOF:
        AT_BOT(FALSE);
        dp->com.cdb[0] = 0x11;  /* (O) Space */
        dp->com.cdb[1] = 0x01;  /* Advance by File Mark */
        dp->com.cdb[2] = 0;
        dp->com.cdb[3] = 0;
        dp->com.cdb[4] = 1; /* # File Marks */
        dp->com.cdb[5] = 0;
        dp->com.reading = FALSE;
        break;
    case WEOF:
        AT_BOT(FALSE);
        dp->com.cdb[0] = 0x10;  /* (M) Write File Marks */
        dp->com.cdb[1] = 0;     /* [FJE] 0x00 for Viper, was 0x01 */
        dp->com.cdb[2] = 0;
        dp->com.cdb[3] = 0;
        dp->com.cdb[4] = 1;
        dp->com.cdb[5] = 0;
        dp->com.reading = FALSE;
        break;
    case FLUSH:
        dp->com.cdb[0] = 0x10;  /* (M) Write File Marks */
        dp->com.cdb[1] = 0;
        dp->com.cdb[2] = 0;
        dp->com.cdb[3] = 0;
        dp->com.cdb[4] = 0;     /* Zero File Marks means Flush */
        dp->com.cdb[5] = 0;
        dp->com.reading = FALSE;
        break;
    default:
DEBUG_ON( VIO_DEFAULT )
        printf(" ignored special.com [in start()]:  %x\n", com);
DEBUG_OFF( VIO_DEFAULT )
    }
    dp->com.card = sdcard(dp - cttab[0]);
    dp->com.unit = sdunit(dp - cttab[sdcard(dp - cttab[0])]);
    dp->com.intr = ihandle;
    dp->state = st;
    sdqueue(&dp->com);

DEBUG_ON( VIO_NON_IO|VIO_IO )
    printf("Command %d (%s)\n", special.com, Commands[special.com]);
DEBUG_OFF( VIO_NON_IO|VIO_IO )

DEBUG_ON( VIO_GENERAL )
    dump_sdcom(" after sdqueue in start()", &dp->com);
DEBUG_OFF( VIO_GENERAL )
}

static int      setresidue(uchar sense[], struct buf * bp)
{
    if (sense[0] == 0xF0) {
        uint r = *(ulong *) & sense[3] * BSIZE;

        if (r <= bp->b_bcount)
            bp->b_resid = r;
    }
    return (bp->b_resid);
}

static int      report(struct ct * dp)
{
    uint            s;

    printf("%s: tape ct%d: ", sdhardwarename(dp->com.card),
           dp->com.card * SDUNITS + dp->com.unit);
    if (!dp->com.okay)
        printf("SCSI timeout\n");
    else
        switch (dp->com.status) {
        case 0:
            s = dp->sense[2] & 0x0F;
            printf("%s (", sensekey(s));
            printbits(dp->sense[2], "FM EOMILI");
            printbits(dp->sense[8], "   CNI   WRPeomUDEBNLFIL");
            printbits(dp->sense[9], "      NDT   BOMBPE");
            printf("%d)\n", *(ushort *) (dp->sense + 10));
            break;
        case 8:
            printf("busy\n");
            break;
        default:
            printf("unknown error\n");
        }
}

static int      printbits(int bits, char *names)
{
    while (*names) {
        if (bits & (1 << 7)) {
            putchar(names[0]);
            putchar(names[1]);
            putchar(names[2]);
            putchar(' ');
        }
        bits <<= 1;
        names += 3;
    }
}
